Program makemap;
{
Create a map on one level using input net from MAKENODE
}

uses Dukedefs,Dukeutil,Duketrig;

Const Version='1.00';
      issue_date='31-Aug-98 ';
      program_name='MAKEMAP1';
      default_template='TEMPLATE.MAP';
      default_node='NODES.TXT';
      author_email='howell@anglianet.co.uk';
      author_name='Jeff Howell ';

Procedure ShowHelp;
{ Gives simple help screen }
begin
   Writeln('Usage: ',program_name,' outfile [nodefile [templatefile]]');
   Writeln('- outfile = Map file to be created.');
   Writeln('  File name must be specified, default extension is .MAP.');
   Writeln('- nodefile = Map topology file created by MAKENODE.EXE.');
   Writeln('  Default is ',default_node,', default extension is .TXT.');
   Writeln('- templatefile = Tiny map file to define decor of caverns.');
   Writeln('  Default is ',default_template,', default extension is .MAP.');
   Writeln(author_name,issue_date,author_email);
   halt;
end;

Const pixels_per_metre = 330;
      height_pix_per_metre = 6000;
      corridor_width : longint = round(2.0*pixels_per_metre);
      corridor_height : longint = round(2.4*height_pix_per_metre);

{*******************************************************************}
{*                            Rooms                                *}
{*******************************************************************}

Var Texture_scale: longint;  { common scale buffer }

Procedure Set_texture_scale (w: integer);
{ sets scale product of wall w at value in scale }
var size,repeats : longint;
begin
   size:= wall_size(w);
   with wall[w]^ do
      if size<=0 then
         xrepeat:=1
      else if Texture_scale<=0 then
         xrepeat:=255
      else begin
         repeats:= (size shl 8) div Texture_scale;
         if repeats>255 then
            xrepeat:=255
         else
            xrepeat:=repeats;
      end;
end;

Procedure Get_texture_scale (w: integer);
{ retrieves scale product of wall w }
begin
   with wall[w]^ do
      if xrepeat=0 then
         Texture_scale:=65536
      else
         Texture_scale:=(wall_size(w) shl 8) div xrepeat;
end;

Procedure Scale_texture(w1,w2: integer);
{ procedure to correct textures of walls for size }
{ by adjusting no of repeats of w2 to match model in w1 }
begin
   Get_texture_scale(w1);
   Set_texture_scale(w2);
end;

procedure new_room(x1,y1,x2,y2,z1,z2: longint; model: integer);
{ create a rectangular room with 4 walls using style of model }
{ note units of x,y are 3mm, max +-65536; units of z are 0.3 mm no limit }
var s,w,w_model: integer;

Procedure MyFirstWall(x0,y0: longint; var w_new: integer);
{ subprocedure of new_room creates a wall but no links }
begin
   NewWall(w_new,w_model);
   with wall[w_new]^do begin
      x:=x0;
      y:=y0;
      nextwall:=-1;     { just in case we have a duff template .. }
      nextsector:=-1;   { .. force all walls to single sided      }
   end;
   inc(sector[s]^.wallnum);
end;

procedure MyNextWall(x0,y0: longint);
{ subprocedure of new_room. adds a wall in sequence }
{ entry with w pointing to previous wall so we can set point2 new wall }
var w_new: integer;
begin
   Get_texture_scale(w_model);         { scale for last wall }
   w_model:=wall[w_model]^.point2;     { step round to this wall model }
   MyFirstWall(x0,y0,w_new);           { make wall }
   wall[w]^.point2:=w_new;             { last wall is now set up }
   Set_texture_scale(w);               { restore texture of last wall }
   w:=w_new;
end;

type alignment_type = (vertical,horizontal);
var alignment : alignment_type;

begin { new_room }
{ determine model offset and starting wall }
{ note model must have 4 walls only }
   w_model:=sector[model]^.wallptr;
   alignment:=alignment_type((x2-x1)>(y2-y1));

{ create new sector }
   NewSector(s,model);
   with sector[s]^ do begin
      ceilingz:=z1;
      floorz:=z2;
      wallnum:=0;
   end;

{ create walls of vertical section .. }
   if alignment=vertical then begin
      MyFirstWall(x1,y1,w);
      Sector[s]^.wallptr:=w;
      MyNextWall(x2,y1);
      MyNextWall(x2,y2);
      MyNextWall(x1,y2);

{ .. or horizontal section }
   end else begin
      MyFirstWall(x2,y1,w);
      Sector[s]^.wallptr:=w;
      MyNextWall(x2,y2);
      MyNextWall(x1,y2);
      MyNextWall(x1,y1);
   end;

{ link last wall back to first one }
   wall[w]^.point2:= sector[s]^.wallptr;
   Scale_texture(w_model,w);
end;

Procedure New_Room_at_Random(x1,y1,model: integer);
{ create a room at specified coarse grid position }
{ give it a random shift and size }

var x,y,dx,dy,z,dz: longint;
begin
   x:=longint(x1)*6000-30000+Random(2000);
   dx:=1500+random(2500);
   y:=longint(y1)*6000-30000+Random(2000);
   dy:=1500+random(2500);
   z:=-corridor_height div 2;
   dz:=corridor_height;
   New_room(x,y,x+dx,y+dy,z-random(corridor_height),z+dz,model);
end;

{*******************************************************************}
{*                            Corridors                            *}
{*******************************************************************}

Procedure Check_facing_walls(a,b: integer; var p: real);
{ computes the cost of joining two walls a,b with a corridor }
{ the cost is lowest for a short corridor normal to both walls }
{ ignore walls which are too narrow for a corridor }
{ and which are already double-sided }
var xa,xb,ya,yb: longint;
    a1,a2: real;
begin
  p:=1e6;
  if wall_size(a)<corridor_width then exit;
  if wall_size(b)<corridor_width then exit;
  if wall[a]^.nextsector>=0 then exit;
  if wall[b]^.nextsector>=0 then exit;

  midpoint(a,xa,ya);
  midpoint(b,xb,yb);
  p:=separation(xa-xb,ya-yb);
  a1:=point_angle(a,xb,yb)-1024;
  a2:=point_angle(b,xa,ya)-1024;
  p:=p+sqr(a1)+sqr(a2);
end;

Procedure Insert_corridor_section(var w0: integer; secnum: integer);
{ splits a wall into three parts with a corridor width section }
{ in the middle. Returns pointer to middle section }
{ increases wall count of master sector }
var w1,w2,angle: integer;
    side1: longint;
begin
{ check we have enough wall to split }
   if wall_size(w0)<corridor_width then exit;
   side1:=round((wall_size(w0)-corridor_width) * random);

{ grab current scale }
   Get_texture_scale(w0);

{ create new wall which will be portal to corridor }
   NewWall(w1,w0);
   with wall[w1]^ do extrapolate(w0,side1,x,y);
   wall[w0]^.point2:=w1;
   with sector[secnum]^ do inc(wallnum);

{ create new wall which will be on other side of portal }
   NewWall(w2,w1);
   with wall[w2]^ do extrapolate(w1,corridor_width,x,y);
   wall[w1]^.point2:=w2;
   with sector[secnum]^ do inc(wallnum);

{ restore all textures }
   Set_texture_scale(w0);
   Set_texture_scale(w1);
   Set_texture_scale(w2);

{ return wall pointer to middle section }
   w0:=w1;
end;

Procedure link_walls(wtop,wbot,model: integer);
{ create a four sided sector s with new walls common to wtop and wbot }
{ for commentary, we assume wtop is north and wbot is south }

var new_s: integer;

Procedure check_angle(point,split: integer);
{ sub_procedure of link_walls }
{ check the angle at the end of 'point'. If less than about 60 degrees }
{ split the wall 'split' and rotate the split point to increase the angle }
{ then retexture the new wall section }

const min_angle = 400; { 512 is normal }
var a, extra_w : integer;
    size: longint;

begin
   a:=normalise_angle(angle(point)-angle(wall[point]^.point2))-1024;
   if a>min_angle then exit;
   size:=wall_size(split);
   if size<corridor_width then exit;

{ insert new node in wall }
   NewWall(extra_w,split);
   inc(sector[new_s]^.wallnum);
   if point=split then

{ allocate links and x,y position for start point of self split wall }
      with wall[extra_w]^ do begin
         point2:=wall[split]^.point2;
         wall[split]^.point2:=extra_w;
         extrapolate(extra_w,size-corridor_width,x,y);
         rotate(x,y,min_angle-a,point2);
      end

{ ditto for start point of separately split wall }
      else with wall[split]^ do begin
         wall[extra_w]^.point2:=split;
         wall[point]^.point2:=extra_w;
         extrapolate(split,corridor_width,x,y);
         rotate(x,y,a-min_angle,extra_w);
      end;

{ retexture extra wall }
   set_texture_scale(extra_w);
end;

var w0,w1,w2,w3: integer;
begin
{ new sector and first wall }
   NewSector(new_s,model);
   with sector[new_s]^ do begin
     ceilingz:= - corridor_height div 2;
     floorz:=  corridor_height div 2;
     NewWall(w0,wallptr);  { wall model is first wall of sector model }
     wallptr:=w0;
     wallnum:=4;    { assume a 4 wall corridor section }
   end;

{ start and finish of east wall }
   with wall[w0]^ do begin
      x:=wall[wtop]^.x;
      y:=wall[wtop]^.y;
      nextsector:=-1;    { tidy up east wall definition }
      nextwall:=-1;
      cstat:=cstat and (-w_hitscan);
      NewWall(w1,w0);    { copy east wall for south wall }
      point2:=w1;
   end;

{ start and finish of west wall }
   NewWall(w2,w0);       { copy tidied east wall for west wall }
   with wall[w2]^ do begin
      x:=wall[wbot]^.x;
      y:=wall[wbot]^.y;
      NewWall(w3,w0);
      point2:=w3;
   end;

{ link southeast corner }
   with wall[wall[wbot]^.point2]^ do begin
      wall[w1]^.x:=x;
      wall[w1]^.y:=y;
   end;

   with wall[w1]^ do begin
      point2:=w2;
      nextwall:=wbot;
   end;

{ link northwest corner }
   with wall[wall[wtop]^.point2]^ do begin
      wall[w3]^.x:=x;
      wall[w3]^.y:=y;
   end;

   with wall[w3]^ do begin
      point2:=w0;
      nextwall:=wtop;
   end;

{ prepare for retexture east and west walls }
   Get_texture_scale(sector[model]^.wallptr);

{ check that corridor angle is sufficiently close to 90 degrees }
{ if any angle is less, split the wall then rotate the new point }
   check_angle(w0,w0);   { SE }
   check_angle(w1,w2);   { SW }
   check_angle(w2,w2);   { NW }
   check_angle(w3,w0);   { NE }

   Set_texture_scale(w0);
   Set_texture_scale(w2);

{ finish links into north and south rooms so game can enter corridor }
   with wall[wtop]^ do begin
     nextwall:=w3;
     nextsector:=new_s;
   end;
   with wall[wbot]^ do begin
      nextwall:=w1;
      nextsector:=new_s;
   end;
end;

Procedure Find_best_pairing(r1,r2: integer; var w1,w2: integer);
{ identify the two facing walls of each room }
{ by searching every pairing of every wall of two rooms }

var g1,g2: integer;
    pmin: real;

Procedure Check_wall_pair;
{ subprocedure of find_best_pairing }
{ updates best potential two walls to join }
var p: real;
begin
   Check_facing_walls(g1,g2,p);
   If p>pmin then exit;
   pmin:=p;
   w1:=g1;
   w2:=g2;
end;

begin { Find_best_pairing }
   pmin:=1e12;
   g1:=sector[r1]^.wallptr;
   repeat
      g2:=sector[r2]^.wallptr;
      repeat
        check_wall_pair;
        g2:=wall[g2]^.point2;
      until g2=sector[r2]^.wallptr;
      g1:=wall[g1]^.point2;
   until g1=sector[r1]^.wallptr;
end;

Procedure Backward_link(my_wall, my_sector: integer);
{ create a link so that game can walk through given wall into sector }
var neighbour: integer;
begin
   neighbour:= wall[my_wall]^.nextwall;
   with wall[neighbour]^ do begin
      nextsector:=my_sector;
      nextwall:=my_wall;
   end;
end;

Procedure Join_rooms_with_corridor(room1,room2,model: integer);
{ given two adjacent rooms, create a corridor between them }

var g1,g2,w1,w2,corridor_sector: integer;
    pmin: real;

begin  { join_rooms_with_corridor }
{ select facing wall pair }
   Find_best_pairing(room1,room2,w1,w2);

{ create extra walls on each side of corridor entry point }
   Insert_corridor_section(w1,room1);
   Insert_corridor_section(w2,room2);

{ create a room which will become the corridor }
   Link_walls(w1,w2,model);

{ create wall and sector pointers out from corridor }
{ note that outside of w1 and w2 point into corridor sector }
   Backward_link(w1,room1);
   Backward_link(w2,room2);
end;

{*******************************************************************}
{*                            Sprites                              *}
{*******************************************************************}

Procedure wall_sprite(e: integer);
{ locate sprite e on any one_sided wall }
{ try for largest walls first  - mark sprite for delete if not OK }
{ 30-8-98 random location along wall and force inside sector }

var best_width,sprite_pos,length_w: longint;
    w,w0: integer;
begin
   w0:=sector[sprite[e]^.sectnum]^.wallptr;
   w:=w0;
   best_width:=corridor_width;

{ consider each wall in turn until one is found that meets all criteria }
   repeat
       with wall[w]^ do begin
         if nextwall=-1 then
            if random<0.25 then begin
              length_w:=wall_size(w);
              if length_w>best_width then

{ found wall so centre sprite in wall and angle points into room, then exit }
{ give it a tiny bit of extra rotation to guarantee the sprite is in sector }
                with sprite[e]^ do begin
                  sprite_pos:=round((length_w-best_width)*random+best_width/2);
                  extrapolate(w,sprite_pos,x,y);
                  rotate(x,y,1,w);
                  ang:=normalise_angle(angle(w)+512);
                  exit;
                end;
             end;
         w:=point2;
      end;

{ counter makes sure we are not locked in a loop forever }
      dec(best_width);
   until best_width<=0;

{ not successful - mark for deletion }
   sprite[e]^.sectnum:=-1;
end;

procedure free_standing_sprite(e: integer);
{ locate a free standing (facing) sprite anywhere in the room }
{ but ensure it does not block the doorways }
var xm,ym: longint;
    counter,w: integer;

begin
   with sprite[e]^ do begin

{ find a spot for this sprite inside the room }
     mid_sector(sectnum,xm,ym);
     counter:=0;
     repeat
        x:=xm+random(5000)-2500;
        y:=ym+random(5000)-2500;
        inc(counter);
     until inside(sectnum,x,y) or (counter>1000);

{ if not placed, set sprite for deletion by squeeze_sprites }
     if not inside(sectnum,x,y) then begin
        sectnum:=-1;
        exit;
     end;

{ ensure that sprite does not block door }
     w:=sector[sectnum]^.wallptr;
     repeat
        if wall[w]^.nextwall>-1 then begin
        { found a wall that backs onto another wall, = a portal }
              midpoint(w,xm,ym);
              if separation(x-xm,y-ym) < (corridor_width *3 div 4) then begin

{ set sprite for deletion by squeeze_sprites }
                 sectnum:=-1;
                 exit;
               end;
         end;
         w:=wall[w]^.point2;
     until w=sector[sectnum]^.wallptr;
   end;
end;

Procedure add_sprite(sprite_num, sec_num: integer);
{ add a sprite into sector according to type of sprite }
var z_height,z_drop: longint;
    hanging_down: boolean;
begin
{ establish vertical position of sprite in template room }
{ remember +ve z is going down }
   with sprite[sprite_num]^ do
     with sector[sectnum]^ do begin
        z_height:=z-floorz;
        z_drop:=ceilingz-z;
     end;
     hanging_down:= z_height < (z_drop*2);

{ create new sprite at correct height }
   newsprite (sprite_num,sprite_num);
   with sprite[sprite_num]^ do begin
       sectnum:=sec_num;
       if hanging_down then
          z:=sector[sec_num]^.ceilingz-z_drop
       else
          z:= z_height+sector[sec_num]^.floorz;
       if (cstat and e_floor)>0 then mid_sector(sectnum,x,y)
       else if (cstat and e_wall)>0 then wall_sprite(sprite_num)
       else free_standing_sprite(sprite_num);
   end;
end;

Procedure Squeeze_sprites;
{ removes all null sprites from sprite array }
var s,d : integer;
begin
   if numsprites=0 then exit;
   d:=0;
   for s:=0 to numsprites-1 do begin
      if sprite[s]^.sectnum>=0 then begin
         if d<s then sprite[d]^:=sprite[s]^;
         inc(d);
      end;
   end;
   numsprites:=d;
end;

{*******************************************************************}
{*                            Main program                         *}
{*******************************************************************}

Const max_styles = 20; { limit of number of room styles }
      max_decor = 50; { max sprites per room style }

var f: text;
    Sectors_in_template : integer;
    Style_start, Style_count: array[0..max_styles] of integer;
    Last_room_style, Last_room: integer;

Procedure Get_num(var n: longint);
{ error checked file read }
begin
   {$I-}
   Readln(f,n);
   {$I+}
   If (IOResult<>0) then Error('invalid data read from node file');
end;

Procedure Get_six_nums(var a,b,c,d,e,g: integer);
{ error checked file read of 6 integers }
begin
  {$I-}
  Readln(f,a,b,c,d,e,g);
  {$I+}
  If IOResult<>0 then Error('invalid data read from node file');
end;

Function Model (style: integer): integer;
{ converts style number to a pointer to a sector in template }
begin
   Model := Style mod sectors_in_template;
end;


Procedure Make_all_rooms (var room_style: integer);
{ create all rooms identified in net list read from open file f }
{ build up lists of rooms in each room style (contiguous sectors) }
{ returns last room style used }
{ this verssion does not use zgrid, it is included for compatibility }

var i, xgrid, ygrid, zgrid, dummy : integer;
    numnodes: longint; { for compatibility with Get_num }

begin
   room_style:= -2;
   While not Eof(f) do begin

      Inc(Room_style,2);
      If (room_style+1)>=max_styles then Error('too many groups in net list');
      Style_start[room_style]:=NumSectors;
      Get_num(numnodes);

      For i:=1 to numnodes do begin
         If Eof(f) then Error('unexpected end of node file');
         Get_six_nums(xgrid,ygrid,zgrid,dummy,dummy,dummy);
         New_room_at_random(xgrid,ygrid,Model(room_style));
      end;

      Style_count[room_style] := NumSectors - Style_start[room_style];

   end;
end;

Procedure Check_join_rooms(room1,room2,style: integer);
{ procedure to check validity of source file }
begin
  if room1=room2 then
     Error('cannot link a room to itself');
  if (room1<style_start[0]) or (room2<style_start[0])
  or (room1>Last_room) or (room2>Last_room) then
     Error('link to non-existent room');
  Join_Rooms_With_Corridor (Room1,Room2,Model(style));
end;

Procedure  Connect_all_rooms;
{ connect rooms by arcs in net list read from open file f }
{ keep track of rooms in Cursector starting from first one in list }
{ odd style numbers are template corridors }

var i,j,dummy,Room_sect,c_sector,c_style : integer;
    numnodes: longint;
    arc: array [1..3] of integer;

begin
   Room_sect:=Style_Start[0];
   c_style:=-1;
   While not Eof(f) do begin

      Inc(c_style,2);
      Style_Start[c_style]:=numsectors;
      Get_num(numnodes);

      For i:=1 to numnodes do begin
         If Eof(f) then Error('unexpected end of node file');
         Get_six_nums(dummy,dummy,dummy,arc[1],arc[2],arc[3]);
         for j:=1 to 3 do
           if arc[j]>0 then
             Check_Join_Rooms (Room_sect,arc[j]+Style_Start[0],c_style);
         inc(Room_sect);
      end;

      Style_count[c_style]:= numsectors - Style_start[c_style];

   end;
end;

Procedure Decorate_all_rooms(last_style: integer);
{ decorate rooms according to their net list grouping }
var  num_sprite_types: integer;

Procedure Decorate_in_style(my_rooms,my_sprites: integer);
{ Sub_procedure of Decorate_all_rooms }
{ select sprites of style my_sprites for rooms of style my_rooms }
var i,j,s_sprites,n_sprites : integer;
    sprite_list : array [0..max_decor-1] of integer;

Procedure swap_integer( var a,b: integer);
var c: integer; begin c:=a; a:=b; b:=c; end;

begin
{ count number of sprites in this style - if none, quit }
   s_sprites:=0;
   for i:=0 to num_sprite_types-1 do
     if sprite[i]^.sectnum=my_sprites then inc(s_sprites);
   if s_sprites=0 then exit;

{ ensure there is at least one of each sprite and one sprite per room }
   n_sprites:=s_sprites;
   if n_sprites<style_count[my_rooms] then
      n_sprites:=style_count[my_rooms];
   if n_sprites>max_decor then
      n_sprites:=max_decor;

{ build list of sprite pointers }
   j:=0;
   while j<n_sprites do
      for i:=0 to num_sprite_types-1 do
         if sprite[i]^.sectnum=my_sprites then
            if j<n_sprites then begin
               sprite_list[j]:=i;
               inc(j);
            end;

{ randomise sprite list }
   for i:=n_sprites-1 downto 1 do
     swap_integer(sprite_list[random(i)],sprite_list[i]);

{ assign sprites to rooms until list is used up }
   for i:=0 to n_sprites-1 do
      Add_sprite(sprite_list[i],
                 Style_start[my_rooms]+(i mod style_count[my_rooms]));

{ if there are more than 2 sprite styles add bonus sprites }
   if s_sprites<2 then exit;
   if n_sprites=s_sprites then exit;
   for i:=1 to n_sprites div 2 do
      Add_sprite(sprite_list[random(n_sprites)],
                 Style_start[my_rooms]+random(style_count[my_rooms]));
end;

var i: integer;

begin {decorate_all_rooms}
{ note end of sprite list }
   num_sprite_types:=numsprites;
{ process each room style in turn }
   for i:=0 to last_style do Decorate_in_style(i,model(i));
end;


Procedure Add_finish_sprite (sect: integer);
{ create a Nuke sprite to terminate the game }
var sprite_num: integer;
begin
   newsprite (sprite_num,0);
   with sprite[sprite_num]^ do begin
         sectnum:=sect;
         z:= sector[sect]^.floorz - corridor_height*3 div 4; { head height }
         cstat:= e_wall+e_centre;
         picnum:= 142; { the nuke button }
         xrepeat:=15;  { these make it about 20 inches across }
         yrepeat:=15;
         shade:=0;
         pal:=0;
         xoffset:=0;
         yoffset:=0;
         hitag:=0;      { these convert it to end game when hit }
         lotag:=32767;
     end;
{ stick it on a suitable wall }
   wall_sprite(sprite_num);
end;

Procedure Start_player_in_sector(n: integer);
{ set up starting position for single player }
begin
   with sector[n]^ do begin
      mid_sector(n,startposx,startposy);
      startposz:=floorz;
      startang:=point_angle(wallptr,startposx,startposy);
      startsectnum:=n;
   end;
end;

Var i: integer;
    template_fn, node_fn, output_fn: string;

begin
   Writeln('Duke Nukem 3D auto mapper v'+Version);

{ load default filenames }
   template_fn:=default_template;
   node_fn:=default_node;
   If ParamCount=0 then ShowHelp;
   output_fn:=ParamStr(1);
   if ParamCount>1 then node_fn:=ParamStr(2);
   If Paramcount>2 then template_fn:=ParamStr(3);

{ Load template file }
   LoadNewFile(template_fn);

{ check template file }
   Sectors_in_template := Numsectors;
   If Sectors_in_template<2 then Error('Template map needs at least 2 sectors');

{ open net file }
   Parse_filename(node_fn,'TXT');
   Assign(f,node_fn);
   {$I-}  Reset(f); {$I+}
   If IOResult<>0 then Error('node file not found: '+node_fn);

   Get_num(RandSeed);
   Make_all_rooms(Last_room_style);
   Last_room:= Style_start[Last_room_style]+Style_count[Last_room_style]-1;

   Reset(f);
   Readln(f);
   Connect_all_rooms;
   Close(f);

   Decorate_all_rooms(Last_room_style+1);

   Start_player_in_sector(style_start[0]);
   Add_finish_sprite( Last_room );

   Writeln('sectors = ',numsectors);
   Writeln('walls = ',numwalls);
   Writeln('sprites = ',numsprites);
   Writeln('Single player start in sector ',Style_start[0]);
   Writeln('Finish in sector ',Last_room);

   SaveNewFile(output_fn);
end.




